//address.cpp

   /*-----------------------------------------------------------------------*\
   | OWLSock For Windows v1.0                                                |
   --------------------------------------------------------------------------|
   | Written By:  Paul Pedriana                                              |
   | Date:        May 7, 1995.                                               |
   | Copyright:   Copyright (c) 1995 by Paul Pedriana.  All Rights Reserved. |
   | UserID(s):   70541,3223                                                 |
   |              70541.3223@compuserve.com                                  |
   --------------------------------------------------------------------------|
   | OWLSock is a Winsock C++ class library for OWL and Windows (16 and 32). |
   | It makes the use of Winsock much simpler and organizes Winsock's        |
   | functionality into logical classes.  See owlsock.doc and the owlsock    |
   | source files for documentation on how to use OWLSock.                   |
   \*-----------------------------------------------------------------------*/

#include "address.hpp"

SocketAddress::SocketAddress(){
//Only the family is specified.  The rest of the data will currently be undefined.
   sa_family = AF_UNSPEC;
}

SocketAddress::SocketAddress(sockaddr& newSockaddr){
//argument is a sockaddr in network byte ordering.
   operator=(newSockaddr);
}

SocketAddress::SocketAddress(SocketAddress& newSocketAddress){
//The argument address should be in network byte ordering.
   operator=(newSocketAddress);
}

SocketAddress::SocketAddress(u_short nNewFamily, char* chData, short nDataLength){
//The argument address should be in network byte ordering.
   SetAddress(nNewFamily, chData, nDataLength);
}

SocketAddress& SocketAddress::operator =(const sockaddr& newSockaddr){
//The argument address should be in network byte ordering.
   SetAddress(newSockaddr.sa_family, newSockaddr.sa_data, 14);
   return *this;
}

int operator ==(const SocketAddress& address1, const SocketAddress& address2){
//Both addresses should be in the same byte ordering as each other.
//We just happen to know for a fact that the sockaddr's have same data as SocketAddress's.
   int i;

   if(address1.sa_family != address2.sa_family)
      return 0;
   for(i=0; i<14; i++){
      if(address1.sa_data[i] != address2.sa_data[i])
         return 0;
   }
   return 1;
}


void SocketAddress::SetAddress(u_short nNewFamily, const char* chData, short nDataLength){
//nNewFamily is AF_INET, etc.
//chData is a pointer to data in the same format as sockaddr.sa_data[].
//nDataLength is the length of chData.  It is limited to 14 bytes in WinSock 1.1.
//The passed data should be in network byte ordering.
   int i;

   sa_family = nNewFamily;
   for(i=0; i<nDataLength && i<14; i++){
      sa_data[i] = chData[i];
   }
}

//-********************************************************************************************

INetSocketAddress::INetSocketAddress():SocketAddress(){
//Empty constructor.  Useful for creating an INetSocketAddress, but delaying the actual
//  assignment of addressing data until later.
   SetAddress(AF_INET, 0, INADDR_ANY); //default values.  Can be changed later. 
}


INetSocketAddress::INetSocketAddress(sockaddr& newSockaddr)
//The argument should be in network byte ordering
   :SocketAddress(newSockaddr)
{
}

INetSocketAddress::INetSocketAddress(u_short nNewPort, u_long lNewAddress, u_short nNewFamily){
//All arguments should be in network byte ordering. nNewFamily is an enumeration, so there is
//  no specific byte ordering.
//Note that lAddress can also be INADDR_ANY, INADDR_LOOPBACK, INADDR_BROADCAST, INADDR_NONE.
//  Also note that in winsock.h (1.1), Microsoft defines these addresses in network byte
//  ordering format, so you don't need to convert to network byte ordering upon passing to this
//  function or any other what expects network byte ordering.   
   SetAddress(nNewFamily, nNewPort, lNewAddress);
}

INetSocketAddress::INetSocketAddress(u_short nNewPort, char* szNewAddress, u_short nNewFamily){
//nNewPort should be passed in network byte ordering.
//szNewAddress is in the format of numerical IP addressing (e.g. "132.212.43.1").  It cannot
//  be in the form of "user@place"
//nFamily is in the form of AF_INET, etc.
   u_long lNewAddress = ConvertAddress(szNewAddress);
   SetAddress(nNewFamily, nNewPort, lNewAddress);
}

INetSocketAddress& INetSocketAddress::operator =(const sockaddr& newSockaddr){
//The argument should be in the same byte ordering (network) as this object.
   SocketAddress::SetAddress(newSockaddr.sa_family, newSockaddr.sa_data, 14);
   return *this;
}

u_long INetSocketAddress::ConvertAddress(char* szAddress){
//This function accepts a string in the form of e.g. 162.132.211.204" and converts it to an
//  IP address (u_long) if possible.
//The returned value will be in network byte ordering. It should be INADDR_NONE if there was an error.
   return inet_addr(szAddress);
}

char* INetSocketAddress::ConvertAddress(u_long lAddress){
//This function accepts a u_long in the form of an IP binary address and converts it to an
//  character string IP address (e.g. "123.213.132.122") if possible.  The lAddress parameter
//  must be given in network byte ordering to work properly.
//Note that this function returns a character pointer to a string allocated by the system.
//  It is imperative that the caller immediately copy this data, and neither modify it nor de-allocate
//  it, etc.  This restriction is imposed by Winsock.
   in_addr tempAddress;
   char*   szAddress;

   tempAddress.s_addr = lAddress;
   szAddress = inet_ntoa(tempAddress);   //sockets call
   return szAddress;
}

short INetSocketAddress::IsAddressDottedDecimal(char* szAddress){
//This function can be used to tell if a character string points to an address in
//  dotted-decimal IP format (e.g. "162.132.211.204") or in name format
//  (e.g. "jimmy_carter@wh.com").  You could call this function if the user
//  typed in a destination address in one of either of the above formats, and you
//  need to know which one it is, so you can know how to convert it to a
//  u_long IP address.  If the address is dotted-decimal, then you can simpoy use
//  the INetSocketAddress::ConvertAddress() function.  Otherwise, you need to use
//  one ofthe HostInfoManager functions.
//The szAddress is a pointer to a string that can be in any format, but most likely, one of
//  either the dotted-decimal or the name formats mentioned above.  This function's
//  job is only to tell you if it is in dotted-decimal format or not.
//The return value is 1 if in dotted-decimal format, and 0 if not.
   unsigned long lReturnValue;

   lReturnValue = inet_addr(szAddress);
   if(lReturnValue == INADDR_NONE)
      return 0;
   return 1;
}

void INetSocketAddress::SetAddress(u_short nNewFamily, u_short nNewPort, u_long lNewAddress){
//This function makes an address out of the necessary IP address components.
//nNewFamily is an enumeration, e.g. AF_INET.
//nNewPort is in network byte ordering, as is lNewAddress.
   SetFamily(nNewFamily);
   SetPort(nNewPort);
   SetNetworkAddress(lNewAddress);
   SetFiller();
}

u_short INetSocketAddress::GetPort(){
//Returns port in network byte ordering
   return ((sockaddr_in*)this)->sin_port;
}

u_long INetSocketAddress::GetNetworkAddress(){
//Returns IP binary address in network byte ordering
   return ((sockaddr_in*)this)->sin_addr.s_addr;
}

u_long INetSocketAddress::GetNetwork(){
//This function returns 0L if the network cannot be determined.
   INetClass theINetClass;
   u_long    lSubnetMask;

   theINetClass = GetClass();
   if(theINetClass == ClassA)
      lSubnetMask = 0xFF000000L;
   else if(theINetClass == ClassB)
      lSubnetMask = 0xFFFF0000L;
   else if(theINetClass == ClassC)
      lSubnetMask = 0xFFFFFF00L;
   else //ClassUnknown
      return 0L;
   return GetNetwork(lSubnetMask);
}

u_long INetSocketAddress::GetNode(){
   INetClass theINetClass;
   u_long    lSubnetMask;

   theINetClass = GetClass();
   if(theINetClass == ClassA)
      lSubnetMask = 0xFF000000L;
   else if(theINetClass == ClassB)
      lSubnetMask = 0xFFFF0000L;
   else if(theINetClass == ClassC)
      lSubnetMask = 0xFFFFFF00L;
   else //ClassUnknown
      return 0L;
   return GetNode(lSubnetMask);
}


void INetSocketAddress::SetPort(u_short nPort){
//This function expects the argument to be in network byte ordering.
   ((sockaddr_in*)this)->sin_port = nPort;
}

void INetSocketAddress::SetNetworkAddress(u_long lAddress){
//This function expects the argument to be in network byte ordering.
   ((sockaddr_in*)this)->sin_addr.s_addr = lAddress;
}

void INetSocketAddress::SetNetworkAddress(char* szAddressDottedDecimal){
   SetNetworkAddress(ConvertAddress(szAddressDottedDecimal));
}



void INetSocketAddress::SetFiller(){
//This is an internal function.  It is merely used to zero out the unused
//  data of the address.
   int i;
   for(i=0; i<8; i++)
      ((sockaddr_in*)this)->sin_zero[i]=0;
}

INetSocketAddress::INetClass INetSocketAddress::GetClass(){
//This function uses WinSock macros to get standard class information.  It knows
//  nothing about subnetting, nor any classes beyond class C.
   if(IN_CLASSA(GetNetworkAddress()))
      return ClassA;
   if(IN_CLASSB(GetNetworkAddress()))
      return ClassB;
   if(IN_CLASSC(GetNetworkAddress()))
      return ClassC;
   return ClassUnknown;
}




