//servmgr.cpp

   /*-----------------------------------------------------------------------*\
   | OWLSock For Windows v1.0                                                |
   --------------------------------------------------------------------------|
   | Written By:  Paul Pedriana                                              |
   | Date:        May 7, 1995.                                               |
   | Copyright:   Copyright (c) 1995 by Paul Pedriana.  All Rights Reserved. |
   | UserID(s):   70541,3223                                                 |
   |              70541.3223@compuserve.com                                  |
   --------------------------------------------------------------------------|
   | OWLSock is a Winsock C++ class library for OWL and Windows (16 and 32). |
   | It makes the use of Winsock much simpler and organizes Winsock's        |
   | functionality into logical classes.  See owlsock.doc and the owlsock    |
   | source files for documentation on how to use OWLSock.                   |
   \*-----------------------------------------------------------------------*/

#include <owl\owlpch.h>
#pragma hdrstop

#include "servmgr.hpp"

//-********************************************************************
DEFINE_RESPONSE_TABLE1(ServiceWindow, TWindow)
   EV_MESSAGE(MSG_SERVICE_NOTIFY, DoNotification),
END_RESPONSE_TABLE;

//A ServiceWindow requires a ServiceManager parent.
ServiceWindow::ServiceWindow(ServiceManager* newServiceManagerParent)
  :TWindow(0, "Service Window"), myServiceManagerParent(newServiceManagerParent)
{
   Attr.Style = 0; //Turn off WS_CHILD (the default) style).
}

//This function merely calls the parent notification function.  Since Windows is designed
//  to only allow Windows to receive system messages, we use a dummy window that does nothing
//  but simulate the concept of Objects receiveing messages, in this case, the ServceManager
//  object.  The ServiceWindow hides this from the user (programmer).
LRESULT ServiceWindow::DoNotification(WPARAM , LPARAM lParam){
   myServiceManagerParent->ServiceCompleted(WSAGETASYNCERROR(lParam));
   return 1L;
}


//-*********************************************************************************
ServiceEntry::ServiceEntry(){
   s_name    = NULL;
   s_aliases = NULL;
   s_port    = 0;
   s_proto   = NULL;
}


ServiceManager::ServiceManager()
   :myWindow(this)
{
   myServiceEntry = (ServiceEntry*)&chServiceBuffer;
   hLastServiceRequest          = 0;
   nLastError                   = 0;
   bLastServiceRequestCompleted = 0;
   nOutstandingServiceRequests  = 0;
   try{
      myWindow.Create();
   }
   catch(...){
      //Invalid Window exception.
   }
}


ServiceManager::~ServiceManager(){
   if(hLastServiceRequest && !bLastServiceRequestCompleted){ //If there is an outstanding request...
      CancelServiceRequest();
   }
}

int ServiceManager::GetService(ServiceEntry*& sEntry, int nPort, const char* szProtocol){
//Implements the blocking 'getservbyport()'.
//sEntry is a pointer to a ServiceEntry*. The caller does NOT allocate a ServiceEntry.
//  It will be set by the system.  DO NOT chanage any data in the returned structure.
//nPort is passed to this function in network byte ordering.
//szProtocol is the protocol name, but may be passed as NULL to mean default or first found.
//The returned sEntry has the port in network byte order.
   sEntry = (ServiceEntry*)getservbyport(nPort, szProtocol); //sockets call
   if(!sEntry){
      nLastError = WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}

int ServiceManager::GetService(ServiceEntry*& sEntry, const char* szName, const char* szProtocol){
//Implements the blocking 'getservbyname()'.
//sEntry is a pointer to a ServiceEntry*. The caller does NOT allocate a ServiceEntry.
//  It will be set by the system.  DO NOT chanage any data in the returned structure.
//The szName argument points to string representing the service name, such as "ftp".  It is
//  generally case sensitive.
//szProtocol is the protocol name, but may be passed as NULL to mean default or first found.
//The returned sEntry has the port in network byte order.
   sEntry = (ServiceEntry*)getservbyname(szName, szProtocol);
   if(!sEntry){
      nLastError = WSAGetLastError();
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}

int ServiceManager::GetServicePort(char* szName, int& nPort, const char* szProtocol){
//This function is blocking.
//Given an input service in "szName", this function fills in nPort with the
//  port (in network order).
//nPort is a reference to an int.  If the call retuns OK, then this will be the port
//  in network ordering.
//szProtocol is the protocol name, but may be passed as NULL to mean default or first found.
   ServiceEntry* tempServiceEntry;
   int nReturnValue;

   nReturnValue = GetService(tempServiceEntry, szName, szProtocol);
   nPort = tempServiceEntry->s_port; //Note that port is already in host byte order.
   return nReturnValue;
}

int ServiceManager::GetServiceName(int nPort, char* szName, const char* szProtocol){
//This function is blocking.
//'szName' needs to be big enough to hold the service's name (N_MAX_SERVICE_NAME).
//nPort must be passed in network byte ordering.
//szProtocol is the protocol name, but may be passed as NULL to mean default or first found.
   ServiceEntry* tempServiceEntry;
   int nReturnValue;

   //Note that because we are calling our own member function, we don't convert
   //  the address from local to network format.
   nReturnValue = GetService(tempServiceEntry, nPort, szProtocol);
   if(nReturnValue == WINSOCK_NOERROR)
      strcpy(szName, tempServiceEntry->s_name);
   return nReturnValue;
}

int ServiceManager::CancelServiceRequest(HANDLE hService){
//Cancels an pending service that equals the hService.  Note that if hService is 0 or
//  unspecified, then this function uses its own current hService.
   int nTempValue;

   if(!hService)
      hService = hLastServiceRequest;
   if(!hService){
      nLastError=WSAEINVAL;
      return WINSOCK_ERROR;
   }
   nTempValue = WSACancelAsyncRequest(hService);   //sockets call
   if(nTempValue == SOCKET_ERROR){
      nLastError = WSAGetLastError();   //sockets call
      return WINSOCK_ERROR;
   }
   nOutstandingServiceRequests--;
   return WINSOCK_NOERROR;
}


int ServiceManager::GetServiceAsync(HANDLE& hService, int nPort, const char* szProtocol){
//This call is non-blocking.  It sets up a callback to its own member window.
//hService is a HANDLE reference.  If the call is successful, it will hold the handle for
//  the asynchrous call.  Note that the caller can save the handle, but also that this
//  class stores the handle for itself as well.
//nPort is passed into this function in network byte ordering.
//Note that due to the design of this class, you cannot have two outstanding Service
//  requests that get notified directly to this class.  You may use the hwnd-specific
//  notification version of this function to manage multiple requests at a time.
//  You may also simply create more than one instance of this class.
//The service is complete when bServiceRequestCompleted is TRUE.  You must look at nLastError in
//  this case to see if there was an error.
   bLastServiceRequestCompleted=0;
   hService = hLastServiceRequest = WSAAsyncGetServByPort(myWindow, MSG_SERVICE_NOTIFY,
                                         nPort, szProtocol, chServiceBuffer, MAXGETHOSTSTRUCT); //sockets call
   if(!hService){
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   nOutstandingServiceRequests++;
   return WINSOCK_NOERROR;
}

int ServiceManager::GetServiceAsync(HANDLE& hService, char* szName, const char* szProtocol){
//This call is non-blocking.  It sets up a callback to its own member window.
//hService is a HANDLE reference.  If the call is successful, it will hold the handle for
//  the asynchrous call.  Note that the caller can save the handle, but also that this
//  class stores the handle for itself as well.
//The 'szName' argument points to a service name, such as "finger".
//szProtocol is the protocol name, but may be passed as NULL or not passed at all to mean
//  default or first found.
   bLastServiceRequestCompleted=0;
   hService = hLastServiceRequest = WSAAsyncGetServByName(myWindow, MSG_SERVICE_NOTIFY,
                                    szName, szProtocol, chServiceBuffer, MAXGETHOSTSTRUCT); //sockets call
   if(!hService){
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   nOutstandingServiceRequests++;
   return WINSOCK_NOERROR;
}


int ServiceManager::GetServiceAsync(TWindow& wndNotify, HANDLE& hService, int nPort, const char* szProtocol,
                                    UINT nMessage, char* chBuffer)
{
//This function notifies the given wndNotify about the completion of the request.
//nMessage is the message that the hwndNotify will receive. It defaults to MSG_SERVICE_NOTIFY,
//  which is defined in the ServiceManager's header file.
//The hService will hold a handle that the caller can use to reference the request on call-back.
//nPort should be passed in network byte ordering.
//The chBuffer is a pointer to buffer that will be filled in with a SERVENT.  Needs to be at least
//  MAXGETHOSTSTRUCT bytes.  If chBuffer is NULL (or not specified), then the ServiceManager's
//  internal buffer will be used.
//szProtocol is the protocol name, but may be passed as NULL to mean default or first found.
//This class will NOT save the hServcice for itself.
//   wParam will be equal to the hService returned.
//   WSAGETSYNCERROR(lParam)   holds an error, if any (0 is OK).
//   WSAGETSYNCBUFLEN(lParam)  hold actual length of buffer
   char* chBufferToUse;

   if(chBuffer)
      chBufferToUse = chBuffer;
   else
      chBufferToUse = chServiceBuffer;

   bLastServiceRequestCompleted=0;
   hService = WSAAsyncGetServByPort(wndNotify, nMessage, nPort,
                                      szProtocol, chBufferToUse, MAXGETHOSTSTRUCT); //sockets call
   if(!hService){
      nLastError = WSAGetLastError();
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}


int ServiceManager::GetServiceAsync(TWindow& wndNotify, HANDLE& hService, char* szName, const char* szProtocol,
                                    UINT nMessage, char* chBuffer)
{
//This function notifies the given wndNotify about the completion of the request.
//nMessage is the message that the hwndNotify will receive. It defaults to MSG_SERVICE_NOTIFY,
//  which is defined in the ServiceManager's header file.
//The hService will hold a handle that the caller can use to reference the request on call-back.
//The chBuffer is a pointer to buffer that will be filled in with a SERVENT.  Needs to be at least
//  MAXGETHOSTSTRUCT bytes.  If chBuffer is NULL (or not specified), then the ServiceManager's
//  internal buffer will be used.
//szName is a pointer to a c string service name, such as "ftp".
//szProtocol is the protocol name, but may be passed as NULL to mean default or first found.
//This class will NOT save the hServcice for itself.
//   wParam will be equal to the hService returned.
//   WSAGETSYNCERROR(lParam)   holds an error, if any (0 is OK).
//   WSAGETSYNCBUFLEN(lParam)  hold actual length of buffer
   char* chBufferToUse;

   if(chBuffer)
      chBufferToUse = chBuffer;
   else
      chBufferToUse = chServiceBuffer;

   bLastServiceRequestCompleted=0;
   hService = WSAAsyncGetServByName(wndNotify, nMessage, szName,
                                     szProtocol, chBufferToUse, MAXGETHOSTSTRUCT); //sockets call
   if(!hService){
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}

void ServiceManager::ServiceCompleted(int nError){
//This function is called whenever an internal window callback is used for an async call.  When
//  this function is called, the bServiceRequestCompleted member will be true (it got set to true
//  right before this function was called).
//You may want to subclass this function.  If you do, you will probably want to call this
//  function as part of the subclassed function.
   if(nError == SOCKET_ERROR)
      nLastError = nError;
   nOutstandingServiceRequests--;
   bLastServiceRequestCompleted=1;
}






