//socksub.cpp

   /*-----------------------------------------------------------------------*\
   | OWLSock For Windows v1.0                                                |
   --------------------------------------------------------------------------|
   | Written By:  Paul Pedriana                                              |
   | Date:        May 7, 1995.                                               |
   | Copyright:   Copyright (c) 1995 by Paul Pedriana.  All Rights Reserved. |
   | UserID(s):   70541,3223                                                 |
   |              70541.3223@compuserve.com                                  |
   --------------------------------------------------------------------------|
   | OWLSock is a Winsock C++ class library for OWL and Windows (16 and 32). |
   | It makes the use of Winsock much simpler and organizes Winsock's        |
   | functionality into logical classes.  See owlsock.doc and the owlsock    |
   | source files for documentation on how to use OWLSock.                   |
   \*-----------------------------------------------------------------------*/

#include <owl\owlpch.h>
#pragma hdrstop

#include "socksub.hpp"
#include "sockmgr.hpp"


//-****************************************************************************
//nMaxPacketSendSize:
//Actually 512 is the absolute guaranteed minimum value.  The WSAData structure has
//  the actual value (>=512).
int DatagramSocket::nMaxPacketSendSize = 512;


DatagramSocket::DatagramSocket():Socket(){
}


DatagramSocket::DatagramSocket(SOCKET& newS):Socket(newS){
}


DatagramSocket::DatagramSocket(SocketAddress& newSocketAddress, int nNewAddressFormat,
    int nNewType, int nNewProtocol)
   :Socket(newSocketAddress, nNewAddressFormat, nNewType, nNewProtocol)
{
}


#ifdef B_USE_DATA_QUEUES
int DatagramSocket::PeekNextSender(SocketAddress& sAddress){
//This function looks into the queue and sees what the address of the next
//  incoming queue item.
//The return value is non-false if an item exists, and false if no item exists.
    QueueItem* peekQueueItem;

    if(myIncomingQueue.GetItemsInContainer()){
       peekQueueItem = myIncomingQueue.Peek();
       sAddress = peekQueueItem->sAddress;
       return 1;
    }
    return 0;
}
#endif //B_USE_DATA_QUEUES



#ifdef B_USE_DATA_QUEUES
int DatagramSocket::AttemptReceiveQueue(DataQueue& queue, int nFlags){
//Reads as much data as possible (until a blocking error) from the driver
//  into the queue.  If the socket is in blocking mode, then the recv call
//  won't return a blocking error; it will simply block.
//We read put the data at the end of the given queue.
//This function returns WINSOCK_ERROR if bad, WINSOCK_NOERROR otherwise.
//  Note that just because a WINSOCK_ERROR is returned, it still is possible
//  that data was read.  For example, the WSAEMSGSIZE error means that data
//  was read, but the buffer wasn't big enough to hold all of it.
//  Also, it is possible that the incoming data is very large and we are able to
//  do a number of reads, but then we get an error.  The first set of reads have
//  put data into our queue, but this function returns WINSOCK_ERROR.
//Attempts to read from the driver that result in a WSAEWOULDBLOCK error are not
//  treated as errors; they are merely treated as though we are being put on hold.
//  This results in this function returning WINSOCK_NOERROR if we get a
//  WSAEWOULDBLOCK error.
//If the socket is in blocking mode, then this call will block, as desired.
//  If the socket is in non-blocking mode, then this function will not block,
//  also as desired.
   int           nCharsReceived;
   QueueItem*    newQueueItem;
   SocketAddress sAddress;
   int           nAddressSize=sizeof(sockaddr);
   char          chBuffer[N_DEF_MAX_READ_BUFFFER_SIZE];

   nCharsReceived = recvfrom(s, chBuffer, N_DEF_MAX_READ_BUFFFER_SIZE, nFlags,
                             &sAddress, &nAddressSize); //sockets call
   if(nCharsReceived == SOCKET_ERROR){
      //It is entirely possible to get a blocking error here.  For example, a
      //  call to recv() could have been made right after the message was
      //  posted by the driver, but before this notification function got called.
      nLastError = WSAGetLastError();
      if(nLastError == WSAEWOULDBLOCK)
         return WINSOCK_NOERROR;
      //Some other error occurred, return from this function.
      return WINSOCK_ERROR;
   }

   //Create a new QueueItem, and pass it the address, buffer, buffer size, and
   //  tell it to copy the buffer and take ownership of its copy.
   newQueueItem = new QueueItem(sAddress, chBuffer, nCharsReceived, 1, 1);
   queue.Put(newQueueItem);

   //We return without an error.  If we are in non-blocking mode, and there is
   //  more data to receive, then the driver will post another message to us.
   return WINSOCK_NOERROR;
}
#endif //B_USE_DATA_QUEUES


int DatagramSocket::Read(char* chData, int& nCharsToRead, SocketAddress& sAddress){
//Reads the Chars into the chData buffer, and removes the data from the queue.
//chData is a pointer to a destination buffer for the data.
//nCharsToRead should be set to the maximum desired read size, which needs to be
//  equal or less to the size of chData.
//The sAddress parameter will get filled with the address of the sender.
//Returns WINSOCK_ERROR if there was an error, WINSOCK_NOERROR otherwise.
//This Read will get the data from the next buffered packet and delete the packet from
//  memory when finished.  Thus if this function is called with a 'nCharsToRead' that
//  is less than the size of the next packet, then only part of the packet will be
//  read and the rest will be lost.
//Upon return, nCharsToRead will be set to the actual number that were read.  If
//  nCharsToRead returns as 0, then no data was read.  If the function return value
//  is WINSOCK_ERROR, then there was a Winsock error, otherwise, the call was
//  successful, even though no data may have been read.
   #ifdef B_USE_DATA_QUEUES
   QueueItem* tempQueueItem;
   int        i;
   int        nError;

   //Try to get a DataQueue item if possible.
   if(myIncomingQueue.GetItemsInContainer()){
      tempQueueItem = myIncomingQueue.Get();
      sAddress      = tempQueueItem->sAddress;
      if(tempQueueItem->nDataLength < nCharsToRead)
         nCharsToRead = tempQueueItem->nDataLength;  //Only read as much as is in the QueueItem.
      for(i=0; i<nCharsToRead; i++)
         chData[i] = tempQueueItem->chData[i];
      return WINSOCK_NOERROR;
   }

   //Try to receive any data if possible.
   nError = AttemptReceiveQueue(myIncomingQueue);
   if(nError == WINSOCK_ERROR)
      return WINSOCK_ERROR;

   //Try to get a DataQueue item if possible.
   if(myIncomingQueue.GetItemsInContainer()){
      tempQueueItem = myIncomingQueue.Get();
      sAddress      = tempQueueItem->sAddress;
      if(tempQueueItem->nDataLength < nCharsToRead)
         nCharsToRead = tempQueueItem->nDataLength;
      for(i=0; i<nCharsToRead; i++)
         chData[i] = tempQueueItem->chData[i];
      return WINSOCK_NOERROR;
   }
   return WINSOCK_NOERROR;

   #else //#ifdef B_USE_DATA_QUEUES
   //Simply call recvfrom() and return if some kind of error is encountered.
   int  nAddressSize=sizeof(sockaddr);
   int  nCharsReceived;

   nCharsReceived = recvfrom(s, chData, nCharsToRead, 0, //Need to change this to allow flags.
                             &sAddress, &nAddressSize); //sockets call
   if(nCharsReceived == SOCKET_ERROR){
      //It is entirely possible to get a blocking error here.  For example, a
      //  call to recv() could have been made right after the message was
      //  posted by the driver, but before this notification function got called.
      nCharsToRead = 0;  //This will be available for the caller to examine.
      nLastError = WSAGetLastError();
      if(nLastError == WSAEWOULDBLOCK)
         return WINSOCK_NOERROR; //This is not considered an "error" under Winsock.
      //Some other error occurred, return from this function.
      return WINSOCK_ERROR;
   }
   nCharsToRead = nCharsReceived;
   return WINSOCK_NOERROR;
   #endif //B_USE_DATA_QUEUES
}



int DatagramSocket::Write(char* chData, int& nCharsToWrite, SocketAddress& thePeerSocketAddress,
   #ifdef B_USE_DATA_QUEUES
                          short bBecomeOwnerOfData, short bCopyData)
   #else //ifdef B_USE_DATA_QUEUES
                          short /*bBecomeOwnerOfData*/, short /*bCopyData*/)
   #endif //B_USE_DATA_QUEUES
{
//Puts the data in the queue and attempts to actually write out the first item in the queue.
//At the end of the function, an attempt to write out the queue is made.  If it fails, well, the
//  data will get sent later, after the system has notified us it is ready.
//Returns WINSOCK_ERROR or WINSOCK_NOERROR.
   #ifdef B_USE_DATA_QUEUES
   QueueItem* tempQueueItem;

   tempQueueItem = new QueueItem(thePeerSocketAddress, chData, nCharsToWrite, bBecomeOwnerOfData, bCopyData);
   myOutgoingQueue.Put(tempQueueItem);  //Adds to end of queue.

   //Trigger an attempt to write.
   return AttemptSendQueue(myOutgoingQueue);

   #else //#ifdef B_USE_DATA_QUEUES
   //Note that bBecomeOwnerOfdata and bCopyData are ignored if we are using DataQueues.
   //Note that thus function may block here if blocking is enabled.
   int nCharsSent;

   nCharsSent = sendto(s, chData, nCharsToWrite, 0, //need to enable flags use.
                         &thePeerSocketAddress, sizeof(sockaddr)); //sockets call
   if(nCharsSent == SOCKET_ERROR){
      //It is entirely possible to get a blocking error here.  For example, a call to
      //  send() could have been made right after the message was posted by the driver,
      //  but before this notification function got called.
      nCharsToWrite = 0;
      nLastError = WSAGetLastError();
      if(nLastError == WSAEWOULDBLOCK)
         return WINSOCK_NOERROR;
      return WINSOCK_ERROR;
   }
   nCharsToWrite = nCharsSent;
   return WINSOCK_NOERROR;
   #endif //B_USE_DATA_QUEUES
}



int DatagramSocket::Write(char* chData, int& nCharsToWrite, short bBecomeOwnerOfData, short bCopyData){
//Simply calls the other Write() function with our latest address.
//Arguments and return values are the same.
   return Write(chData, nCharsToWrite, myPeerSocketAddress, bBecomeOwnerOfData, bCopyData);
}



DatagramSocket::DoReadNotification(const SOCKET& /*socket*/, int nError){
//This function gets called whenever the socket gets a read notification.  This means
//  that data on the port is ready to be read.
//This function doesn't do much with the nError parameter.  It simply doesn't do the
//  read if there is an error value.  According to Winsock documentation, this
//  error may be "any error in winsock.h".
   if(nError){
      nLastError= nError; /*WSAGetLastError(); */
      return WINSOCK_ERROR;
   }

   //This function dan't do anything on its own without data queues.  Without
   //  DataQueues, you can do two things:  1) subclass DatagramSocket and write
   //  your own DoReadNotification() or 2) Have the socket redirect the FD_READ
   //  notification to the window of your choice with Socket::SetNotificationSet()
   //  and SetNotificationWindow().
   #ifdef B_USE_DATA_QUEUES
   return AttemptReceiveQueue(myIncomingQueue);
   #else
   return WINSOCK_NOERROR;
   #endif //B_USE_DATA_QUEUES
}



DatagramSocket::DoWriteNotification(const SOCKET& /*s*/, int nError){
   if(nError){
      nLastError = nError; /*=WSAGetLastError();*/
      return WINSOCK_ERROR;
   }

   //This function dan't do anything on its own without data queues.  Without
   //  DataQueues, you can do two things:  1) subclass DatagramSocket and write
   //  your own DoWriteNotification() or 2) Have the socket redirect the FD_WRITE
   //  notification to the window of your choice with Socket::SetNotificationSet()
   //  and SetNotificationWindow().
   #ifdef B_USE_DATA_QUEUES
   return AttemptSendQueue(myOutgoingQueue);   //Try to send data out of regular queue.
   #else
   return WINSOCK_NOERROR;
   #endif //B_USE_DATA_QUEUES
}



#ifdef B_USE_DATA_QUEUES
int DatagramSocket::AttemptSendQueue(DataQueue& queue, int nFlags){
//This function tries to send as much data from the queue as possible.
//Returns WINSOCK_ERROR or WINSOCK_NOERROR.
   QueueItem* tempQueueItem;
   int        nCharsSent;
   int        nCharsToSend;

   while(1){ //Keep repeating this loop until we are out of data, or a block occurs, or an error occurs.
      if(!queue.GetItemsInContainer())
         return WINSOCK_NOERROR; //Simply nothing to send.
      //We need to do a Peek() instead of a Get() because we won't necessarily be removing
      //  the item from the queue.  We remove it from the queue when we have sent ALL its
      //  information.  For Datagram sockets, we remove the item from queue if any number
      //  of Chars have been sent.
      tempQueueItem = queue.Peek();

      //Try to send what is remaining in the buffer; may be some or all of the original buffer.
      //lPresentPosition starts out to zero for new QueueItems.
      if(tempQueueItem->nDataLength > nMaxPacketSendSize){
         //This is not good.  The user is trying to send datagrams greater than the max size.
         nCharsToSend = nMaxPacketSendSize; //May generate compiler warnings due to conversion issues.
      }
      else
         nCharsToSend = tempQueueItem->nDataLength;
      nCharsSent = sendto(s, tempQueueItem->chData, nCharsToSend, nFlags,
                            &(tempQueueItem->sAddress), sizeof(sockaddr)); //sockets call
      if(nCharsSent == SOCKET_ERROR){
         //It is entirely possible to get a blocking error here.  For example, a call to
         //  send() could have been made right after the message was posted by the driver,
         //  but before this notification function got called.
         //Note that we exit this loop without changing the present QueueItem's data.  Thus,
         //  next time this notification function is called, it will be as if this QueueItem
         //  send hasn't been attempted yet.
         nLastError = WSAGetLastError();
         if(nLastError == WSAEWOULDBLOCK){
            //We simply return.  The current QueueItem is still in the queue for next time.
            //The driver will notify us again later when we can send more data.
            return WINSOCK_NOERROR;
         }
         //Some other error occurred.
         return WINSOCK_ERROR;
      }

      //We are done with this QueueItem. Burn it.
      queue.Get();
      if(nCharsSent < nCharsToSend)
         return WINSOCK_ERROR; //We couldn't send the entire QueueItem.  Not Good At All.
   }
}
#endif //B_USE_DATA_QUEUES



//-******************************************************************************************
StreamSocket::StreamSocket()
   : Socket(), nConnectStatus(nNotConnected)
{
}


StreamSocket::StreamSocket(SOCKET& newS)
   :Socket(newS)
{
   //Important:  Unless there is a way I haven't found yet, you can't tell
   //  what the nConnectStatus is, given just a SOCKET descriptor.
   //Needs to be set somehow:
   //nConnectStatus = ???;
}


//Constructor for a protocol defined by an int.
StreamSocket::StreamSocket(SocketAddress& newSocketAddress, int nNewAddressFormat, int nNewType, int nNewProtocol)
   :Socket(newSocketAddress, nNewAddressFormat, nNewType, nNewProtocol), nConnectStatus(nNotConnected)
{
}


StreamSocket& StreamSocket::operator=(StreamSocket& newStreamSocket){
   Socket::operator=(newStreamSocket);

   #ifdef B_USE_DATA_QUEUES
   CopyQueue(myIncomingOOBQueue, newStreamSocket.myIncomingOOBQueue);
   CopyQueue(myOutgoingOOBQueue, newStreamSocket.myOutgoingOOBQueue);
   #endif //B_USE_DATA_QUEUES

   return *this;
}



#ifdef B_USE_DATA_QUEUES
int StreamSocket::GetLocalWaitingSizeOOB(){
//Very similar to GetTotalReadInLength() except operators on OOB queue.
   return GetLocalWaitingSizeQueue(myIncomingOOBQueue);
}
#endif //B_USE_DATA_QUEUES



#ifdef B_USE_DATA_QUEUES
int StreamSocket::GetLocalWaitingCountOOB(){
//Returns the number of incoming buffers.
//Since this is a stream socket, you usually won't be interested in how many buffers
//  there are, but it may be useful to know under some circumstances.
   return myIncomingOOBQueue.GetItemsInContainer();
}
#endif //B_USE_DATA_QUEUES



#ifdef B_USE_DATA_QUEUES
int StreamSocket::AttemptReceiveQueue(DataQueue& queue, int nFlags){
//Reads as much data as possible (until a block) from the driver.
//We read put the data at the end of the queue.
//This function returns WINSOCK_ERROR if bad, WINSOCK_NOERROR otherwise.
//  Note that just because a WINSOCK_ERROR is returned, it still is possible that
//  data was read.  For example, the WSAEMSGSIZE error means that data was read, but the
//  buffer wasn't big enough to hold all of it.
//  Also, it is possible that the incoming data is very large and we are able to
//  do a number of reads, but then we get an error.  The first set of reads have
//  put data into our queue, but this function returns WINSOCK_ERROR.
//Attempts to read from the driver that result in a WSAEWOULDBLOCK error are not
//  treated as errors; they are merely treated as though we are being put on hold.
//  This results in this function returning WINSOCK_NOERROR if we get a WSAEWOULDBLOCK error.
   int           nCharsReceived;
   QueueItem*    tempQueueItem;
   unsigned long lCharsWaiting;
   int           nCharsToRead;
   char          chBuffer[N_DEF_MAX_READ_BUFFFER_SIZE];

   //Determine the actual number of Chars to read.  We don't want to read too
   //  many Chars if there are a lot waiting.  Also, since the ioctlsocket()
   //  returns a u_long, we must do a safe conversion to int before determining
   //  how much we should read.
   lCharsWaiting = GetDriverWaitingSize();
   if(lCharsWaiting > (long)nMaxReadBufferSize || lCharsWaiting == 0)
      nCharsToRead = nMaxReadBufferSize;
   else
      nCharsToRead = (int)lCharsWaiting; //Could cause a compiler warning, but we have no choice.

   //Try to receive the characters.
   nCharsReceived = recv(s, chBuffer, nCharsToRead, nFlags); //sockets call
   if(nCharsReceived == SOCKET_ERROR){
      //It is entirely possible to get a blocking error here.  For example, a call to
      //  recv() could have been made right after the message was posted by the driver,
      //  but before this notification function got called.
      nLastError = WSAGetLastError(); //sockets call
      if(nLastError == WSAEWOULDBLOCK)
         return WINSOCK_NOERROR;
      //Some other error occurred.
      return WINSOCK_ERROR;
   }
   tempQueueItem = new QueueItem(myPeerSocketAddress, chBuffer, nCharsReceived, 1, 0);
   queue.Put(tempQueueItem);

   //If the above loop exited because nCharsWaiting was zero, then nTempError should be WINSOCK_NOERROR.
   return WINSOCK_NOERROR;
}
#endif //B_USE_DATA_QUEUES



#ifdef B_USE_DATA_QUEUES
int StreamSocket::ReadQueue(DataQueue& queue, char* chData, int& nCharsToRead){
//This function reads from the given queue. It will read any number of buffers from the
//  queue, up till the length of 'nCharsToRead'.
//We first try to satisfy the data request by using the internal queue.  Then we try
//  to read from the driver and check the queue again for data.  The loop looks funny
//  because it is designed to (asynchronous mode):
//     1 try to get the data from the queue if possible (return if any read)
//     2 then try to read from the driver if possible
//     3 then try to get anay data from the queue if possible (return if any read)
//     4 return without any data read.
//'chData' is a pointer to an area of memory to which the data will be copied.
//'nCharsToRead' holds the number of chars max that the caller wants to copy, with a maximum
//  of the size of chData.  Upon return, nCharsToRead will hold the number of Chars
//  actually read.
//Since this is a stream socket, the sending address is always the same.
//Any QueueItems that are completely read are discarded, as you would expect.
//Returns WINSOCK_ERROR or WINSOCK_NOERROR.  If any Chars were read, then they are
//  in nCharsToRead upon return.
   int nPresentBuffer, nCurrentCharReadPosition, nNewCurrentCharReadPosition, nCharsInCurrentBuffer;
   int nCharsLeftToRead, nCharsToReadFromCurrentBuffer, nBuffersInQueue;
   int i, j, jEnd, nError;
   QueueItem* tempQueueItem;
   short bThisLineNotExecutedYet=0;

 while(1){
   nBuffersInQueue = queue.GetItemsInContainer();
   for(nPresentBuffer=nCurrentCharReadPosition=0;
       nCurrentCharReadPosition < nCharsToRead && nPresentBuffer < nBuffersInQueue;
       nPresentBuffer++)
   {
      tempQueueItem = queue.Peek();
      nCharsInCurrentBuffer = tempQueueItem->nDataLength - tempQueueItem->nPresentPosition;
      nCharsLeftToRead = nCharsToRead - nCurrentCharReadPosition;
      if(nCharsLeftToRead  >= nCharsInCurrentBuffer){
         nCharsToReadFromCurrentBuffer = nCharsInCurrentBuffer;  //Read the entire buffer
         queue.Get(); //tempQueueItem already holds the return value, so we don't re-assign it here.
      }
      else{ //The Chars to read are less than the size of the buffer. Also, need to save the buffer.
         nCharsToReadFromCurrentBuffer = nCharsLeftToRead;
         //Don't say 'queue.Get()' here because we will need to save the currrent queue
         //  item in the queue, since there is more data in it to read later.
         //However, we do need to leave a 'note' for ourselves here that tells us where
         //  we last left off the reading.  So next time we start reading from the buffer,
         //  we can start at the appropriate place.  The 'note' is the 'nPresentPosition' member:
         tempQueueItem->nPresentPosition = nCharsToReadFromCurrentBuffer;
      }
      nNewCurrentCharReadPosition = nCurrentCharReadPosition + nCharsToReadFromCurrentBuffer;
      for(j=0, i=tempQueueItem->nPresentPosition, jEnd = nCharsToReadFromCurrentBuffer; j<jEnd; j++, i++)
         chData[nCurrentCharReadPosition+j] = tempQueueItem->chData[i];
      nCurrentCharReadPosition = nNewCurrentCharReadPosition;
   }

   //At this point, 'nCharsToRead' is set to the actual amount that were read, zero or greater.
   if(nCurrentCharReadPosition){ //If any were read, then return. Otherwise, try to receive more data.
      nCharsToRead = nCurrentCharReadPosition;
      return WINSOCK_NOERROR;
   }

   //No characters were in the queue.
   if(bThisLineNotExecutedYet){
      bThisLineNotExecutedYet=1; //We only want to try to receive data from the driver once.
      //Try to receive any data if possible from the driver.
      nError = AttemptReceiveQueue(myIncomingQueue);
      if(nError == WINSOCK_ERROR)
         return WINSOCK_ERROR;
   }
   else{
      break; //break from the while(1) loop.
   }
 }
 nCharsToRead = 0; //None were read.
 return WINSOCK_NOERROR;
}
#endif //B_USE_DATA_QUEUES



int StreamSocket::Read(char* chData, int& nCharsToRead){
//This reads from the already received and queued data.  This data has already
//  been received from the socket driver.
   #ifdef B_USE_DATA_QUEUES
   //First try to get a DataQueue item if possible.
   return ReadQueue(myIncomingQueue, chData, nCharsToRead);

   #else //#ifdef B_USE_DATA_QUEUES
   //Try to receive the characters.
   int nCharsReceived;
   nCharsReceived = recv(s, chData, nCharsToRead, 0); //sockets call //I Need to allow the use of flags here.
   if(nCharsReceived == SOCKET_ERROR){
      //It is entirely possible to get a blocking error here.  For example, a call to
      //  recv() could have been made right after the message was posted by the driver,
      //  but before this notification function got called.
      nCharsToRead = 0; //This will be available for the caller to examine.
      nLastError = WSAGetLastError(); //sockets call
      if(nLastError == WSAEWOULDBLOCK)
         return WINSOCK_NOERROR;
      //Some other error occurred.
      return WINSOCK_ERROR;
   }
   nCharsToRead = nCharsReceived;
   return WINSOCK_NOERROR;
   #endif //B_USE_DATA_QUEUES
}


#ifdef B_USE_DATA_QUEUES
int StreamSocket::WriteQueue(DataQueue& queue, char* chData, int& nCharsToWrite, int nFlags,
                             short bBecomeOwnerOfData, short bCopyData){
//This function puts the data in the queue to be written to the sockets driver.
//At the end of the function, an attempt to write out the queue is made.  If it fails, well, the
//  data will get sent later, after the system has notified us it is ready.
//Returns WINSOCK_ERROR or WINSOCK_NOERROR.
   QueueItem* tempQueueItem;

   tempQueueItem = new QueueItem(myPeerSocketAddress, chData, nCharsToWrite, bBecomeOwnerOfData, bCopyData);
   queue.Put(tempQueueItem);  //Adds to end of queue.

   //Trigger an attempt to write.
   return AttemptSendQueue(queue, nFlags); //In theory, you could also say: DoWriteNotification(s, 0);
}
#endif //B_USE_DATA_QUEUES



int StreamSocket::Write(char* chData, int& nCharsToWrite, int nFlags,
#ifdef B_USE_DATA_QUEUES
                        short bBecomeOwnerOfData, short bCopyData){
#else
                        short /*bBecomeOwnerOfData*/, short /*bCopyData*/){
#endif
   #ifdef B_USE_DATA_QUEUES
   return WriteQueue(myOutgoingQueue, chData, nCharsToWrite, nFlags, bBecomeOwnerOfData, bCopyData);

   #else //#ifdef B_USE_DATA_QUEUES
   int nCharsSent;
   nCharsSent = send(s, chData, nCharsToWrite, nFlags);  //sockets call
   if(nCharsSent == SOCKET_ERROR){
      //It is entirely possible to get a blocking error here.  For example, a call to
      //  send() could have been made right after the message was posted by the driver,
      //  but before this notification function got called.
      nCharsToWrite = 0;
      nLastError = WSAGetLastError(); //sockets call
      if(nLastError == WSAEWOULDBLOCK){
         return WINSOCK_NOERROR;
      }
      return WINSOCK_ERROR;
   }
   nCharsToWrite = nCharsSent;
   return WINSOCK_NOERROR;
   #endif //B_USE_DATA_QUEUES
}


#ifdef B_USE_DATA_QUEUES
int StreamSocket::PeekOOB(char* chData, int& nCharsToRead){
//Works just like the Peek() function but it works on the OOB queue.
   //Read from as many buffers in the queue as possible (999).
   return PeekQueue(myIncomingOOBQueue, chData, nCharsToRead, myPeerSocketAddress, 999);
}
#endif //B_USE_DATA_QUEUES



int StreamSocket::ReadOOB(char* chData, int& nCharsToRead){
//Works just like the Read() function but it works on the OOB queue.
   #ifdef B_USE_DATA_QUEUES
   return ReadQueue(myIncomingOOBQueue, chData, nCharsToRead);

   #else //#ifdef B_USE_DATA_QUEUES
   //Try to receive the characters.
   int nCharsReceived;                    //I Need to allow the use of flags here.
   nCharsReceived = recv(s, chData, nCharsToRead, MSG_OOB); //sockets call
   if(nCharsReceived == SOCKET_ERROR){
      //It is entirely possible to get a blocking error here.  For example, a call to
      //  recv() could have been made right after the message was posted by the driver,
      //  but before this notification function got called.
      nCharsToRead = 0;
      nLastError = WSAGetLastError(); //sockets call
      if(nLastError == WSAEWOULDBLOCK)
         return WINSOCK_NOERROR;
      //Some other error occurred.
      return WINSOCK_ERROR;
   }
   nCharsToRead = nCharsReceived;
   return WINSOCK_NOERROR;
   #endif //B_USE_DATA_QUEUES
}


int StreamSocket::WriteOOB(char* chData, int& nCharsToWrite, int nFlags,
#ifdef B_USE_DATA_QUEUES
                        short bBecomeOwnerOfData, short bCopyData){
#else
                        short /*bBecomeOwnerOfData*/, short /*bCopyData*/){
#endif
//ORs nFlags w MSG_OOB. Otherwise works just like StreamSocket::Write().
   #ifdef B_USE_DATA_QUEUES
   nFlags |= MSG_OOB;
   return WriteQueue(myOutgoingOOBQueue, chData, nCharsToWrite, nFlags, bBecomeOwnerOfData, bCopyData);

   #else //#ifdef B_USE_DATA_QUEUES
   int nCharsSent;
   nFlags |= MSG_OOB;
   nCharsSent = send(s, chData, nCharsToWrite, nFlags);  //sockets call
   if(nCharsSent == SOCKET_ERROR){
      //It is entirely possible to get a blocking error here.  For example, a call to
      //  send() could have been made right after the message was posted by the driver,
      //  but before this notification function got called.
      nCharsToWrite = 0;
      nLastError = WSAGetLastError(); //sockets call
      if(nLastError == WSAEWOULDBLOCK){
         return WINSOCK_NOERROR;
      }
      return WINSOCK_ERROR;
   }
   nCharsToWrite = nCharsSent;
   return WINSOCK_NOERROR;
   #endif //B_USE_DATA_QUEUES
}


StreamSocket::Listen(int nMaxQueuedConnections){
//This function puts this socket into a passive listening mode.
   int nReturnValue;

   nReturnValue = listen(s, nMaxQueuedConnections);
   if(nReturnValue){
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   nConnectStatus = nListening;
   return WINSOCK_NOERROR;
}


StreamSocket::Connect(SocketAddress& addressToConnectTo){
//This function sets myPeerSocketAddress to sAddressToConnectTo then calls Connect().
//See Connect() for more details on it.
   SetPeerSocketAddress(addressToConnectTo);
   return Connect();
}


StreamSocket::Connect(){
//This function uses 'myPeerSocketAddress'; it needs to be set before callign this function.
//The connection attempt (and this function) should return right away, without blocking.
//When we are actually connected, we get a notification from the driver at our
//  'DoConnectNotification()' function.  Upon receiving that notification, the nConnectStatus
//  will get set to 'nConnected'.
//Technically, it is true that a datagram socket can call connect; doing this sets the default address
//  for future send()/recv() calls tha the datagram socket might use.
//  We put it in the StreamSocket class for simplicity, and because our DatagramSocket class
//  already supports its own default address system.
   int nReturnValue;

   nReturnValue = connect(s, &myPeerSocketAddress, sizeof(sockaddr)); //sockets call
   if(nReturnValue){
      nLastError = WSAGetLastError(); //sockets call
      return WINSOCK_ERROR;
   }
   //The following code assumes that the socket is in non-blocking mode.  If the socket was in
   //  blocking mode, then nConnectStatus would be nConnected.  Thus, this code is slightly
   //  bugged and can be patched with a call to select() to determine whether the socket is
   //  really connected or not.  Looked at another way, a blocking socket can be said to be
   //  connected if nConnectStatus is either nConnected or nConnectPending. 
   nConnectStatus = nConnectPending;
   return WINSOCK_NOERROR;
}


StreamSocket::Accept(StreamSocket& socket){
//This function will try to accept a connection with the first connecting peer that
//  is waiting in the queue.  If successful, the StreamSocket& 'socket' will have a
//  valid and connected socket, a proper status of nConnected, and the correct
//  peer socket address.
//The caller will usually call this function in response to an Accept notification.
//  The caller merely needs to create a new StreamSocket and pass it to this
//  function. The default constructor for StreamSocket can be used, as this function
//  fixes up the missing parts.    
   int nReturnValue;

   nReturnValue = Accept(socket.s, socket.myPeerSocketAddress);
   if(nReturnValue == WINSOCK_NOERROR){
      socket.nConnectStatus = nConnected;
      socket.SetSocketStyle(nFamily, nType, nProtocol); //Copy our nFamily, etc. just to make sure.
   }
   return nReturnValue;
}


StreamSocket::Accept(SOCKET& socket, sockaddr& sAddress){
//This function will try to accept a connection with the first connecting peer that is
//  waiting in the queue.  If successful, the 'socket' reference argument will be set to
//  a new connected socket.  The sAddress reference argument will get set to the address
//  of the connecting peer.  The caller of this function may immediately use the
//  new socket with data sends, etc.  Note that the caller may want to flag the socket
//  as connected.  If the 'socket' belongs to a 'StreanSocket', then you can set its
//  nConnectionStatus as nConnected.
//The return value is either WINSOCK_ERROR or WINSOCK_NOERROR.  If there is an error,
//  then nLastError will be set with the appropriate error.
//Note that it is possible that this call could be made when there are no pending
//  socket connecions in the queue.  If this is the case, the call will block
//  if the socket is marked as blocking, and will return WINSOCK_ERROR with
//  WSAEWOULDBLOCK if the socket is marked as non-blocking.
//This function is usually called in response to an accept notification.  A socket is
//  set up as a stream socket and listen() is called.  When a connection is ready, the
//  driver notifies the listening socket with a DoAcceptNotification() call.  See the
//  DoAcceptNotification() call for what to do with it.  You will want to call this Accept()
//  function as a result.
   int nAddressLength = sizeof(sockaddr);

   socket = accept(s, &sAddress, &nAddressLength);  //sockets call
   if(socket == INVALID_SOCKET){
      nLastError = WSAGetLastError();  //sockets call
      return WINSOCK_ERROR;
   }
   return WINSOCK_NOERROR;
}


#ifdef B_USE_DATA_QUEUES
int StreamSocket::AttemptSendQueue(DataQueue& queue, int nFlags){
//If we have data in the queue, then write it out the socket. We try to write as much as possible.
//  Since this is a stream socket, we may write as much as we want (?).
//This function will write as much data as it can until the driver tells us that any more data
//  would cause blocking.
//This function returns WINSOCK_ERROR or WINSOCK_NOERROR.  Check nLastError if there was a return
//  error.
   QueueItem* tempQueueItem;
   int        nCharsSent;
   int        nEndingPosition;

   while(1){
      if(!queue.GetItemsInContainer())
         return WINSOCK_NOERROR; //Simply nothing to send.
      //We need to do a Peek() instead of a Get() because we won't necessarily be removing
      //  the item from the queue.  We remove it from the queue when we have sent ALL its
      //  information.
      tempQueueItem = queue.Peek();

      //This is just a paranoid sanity check:
      //if(tempQueueItem->nPresentPosition >= tempQueueItem->nDataLength){
      //   queue.Get();//Burn this QueueItem.
      //   continue;   //Start again at top of loop.
      //}

      //Try to send what is remaining in the buffer; may be some or all of the original buffer.
      //nPresentPosition starts out to zero for new QueueItems.
      nCharsSent = send(s, tempQueueItem->chData+tempQueueItem->nPresentPosition,
                        tempQueueItem->nDataLength-tempQueueItem->nPresentPosition, nFlags);  //sockets call
      if(nCharsSent == SOCKET_ERROR){
         //It is entirely possible to get a blocking error here.  For example, a call to
         //  send() could have been made right after the message was posted by the driver,
         //  but before this notification function got called.
         //Note that we exit this loop without changing the present QueueItem's data.  Thus,
         //  next time this notification function is called, it will be as if this QueueItem
         //  send hasn't been attempted yet.
         nLastError = WSAGetLastError(); //sockets call
         if(nLastError == WSAEWOULDBLOCK){
            //We simply return.  The current QueueItem is still in the queue for next time.
            //The driver will notify us again later when we can send more data.
            return WINSOCK_NOERROR;
         }
         //Some other error occurred.
         return WINSOCK_ERROR;
      }
      //Otherwise, we may have sent some or all of the characters. Update the present QueueItem
      //  information to reflect this.
      nEndingPosition = tempQueueItem->nPresentPosition + nCharsSent;
      if(nEndingPosition < tempQueueItem->nDataLength){ //If not all of the QueueItem was sent...
         tempQueueItem->nPresentPosition = nEndingPosition;
      }
      else{  //All of the current QueueItem has been sent.
         //We are done with this QueueItem.  Burn it.
         queue.Get();
      }
   }
}
#endif //B_USE_DATA_QUEUES



StreamSocket::DoReadNotification(const SOCKET& /*socket*/, int nError){
//This function gets called whenever the socket gets a read notification.  This means
//  that data on the port is ready to be read.
//This function doesn't do much with the nError parameter.  It simply doesn't do the
//  read if there is an error value.  According to Winsock documentation, this
//  error may be "any error in winsock.h".
   if(nError){
      nLastError= nError; /*WSAGetLastError(); */
      return WINSOCK_ERROR;
   }

   #ifdef B_USE_DATA_QUEUES
      return AttemptReceiveQueue(myIncomingQueue);
   #else
      return WINSOCK_NOERROR;
   #endif //B_USE_DATA_QUEUES
}


StreamSocket::DoWriteNotification(const SOCKET& /*s*/, int nError){
//This function gets called whenever the socket gets a write notification.  This means
//  that data on the port is ready to be written.
   if(nError){
      nLastError = nError; /*=WSAGetLastError();*/
      return WINSOCK_ERROR;
   }

   //This function dan't do anything on its own without data queues.  Without
   //  DataQueues, you can do two things:  1) subclass StreamSocket and write
   //  your own DoWriteNotification() or 2) Have the socket redirect the FD_WRITE
   //  notification to the window of your choice with Socket::SetNotificationSet()
   //  and SetNotificationWindow().
   #ifdef B_USE_DATA_QUEUES
      //The problem with this design right now is that if there is a lot of data in
      //  the regular queue, then the regular queue data sending will saturate
      //  the driver and so OOB sends won't be able to occur.
      int nReturnValue;
      nReturnValue = AttemptSendQueue(myOutgoingQueue);   //Try to send data out of regular queue.
      if(nReturnValue != WINSOCK_NOERROR)
         return nReturnValue;
      return AttemptSendQueue(myOutgoingOOBQueue, MSG_OOB);  //Try to send data out of OOB queue.
   #else
      return WINSOCK_NOERROR;
   #endif //B_USE_DATA_QUEUES
}

StreamSocket::DoOOBNotification(const SOCKET& /*s*/, int nError){
//We get this notification when OOB data is ready to be received on the socket port.
   if(nError){
      nLastError= nError; /*WSAGetLastError();*/
      return WINSOCK_ERROR;
   }

   //This function dan't do anything on its own without data queues.  Without
   //  DataQueues, you can do two things:  1) subclass StreamSocket and write
   //  your own DoOOBNotification() or 2) Have the socket redirect the FD_OOB
   //  notification to the window of your choice with Socket::SetNotificationSet()
   //  and SetNotificationWindow().
   #ifdef B_USE_DATA_QUEUES
      return AttemptReceiveQueue(myIncomingOOBQueue, MSG_OOB);
   #else
      return WINSOCK_NOERROR;
   #endif //B_USE_DATA_QUEUES
}

StreamSocket::DoAcceptNotification(const SOCKET& /*s*/, int /*nError*/){
//We get this notification when a client socket on the network is attempting to connect to us.
   //Code needs to be written to intercept this notification.
   return 0; //We don't do anything.  We let the pending acceptance sit there.
}


StreamSocket::DoConnectNotification(const SOCKET& /*s*/, int nError){
//This means that the connection that we attempted with a server on the network has completed.
//This function gets called sometime after this object makes a connect() attempt.  If the
//  connect attempt was non-blocking, then a notification is posted and we end up here.
//When this function gets called, we our nConnectStatus should be 'nConnecting'.
   if(nError){
      //There was an error, and we cannot connect; this may due to a number of reasons.
      //We turn off our 'nConnectStatus' to 'nNotConnected'.  This is becuase our attempt
      //  to connect failed.
      nConnectStatus = nNotConnected;
      nLastError= nError; /*WSAGetLastError();*/
      return WINSOCK_ERROR;
   }
   nConnectStatus = nConnected;
   return WINSOCK_NOERROR;
}


StreamSocket::DoCloseNotification(const SOCKET& /*s*/, int nError){
//This notification gets called when the socket has been closed.
//We mark the socket as not connected, so that the user may know about it.
//It is important that we read any data that may be waiting in the queue before
//  changing the status of the connection and doing any notification.
   if(nError){
      nLastError= nError; /*WSAGetLastError();*/
      return WINSOCK_ERROR;
   }

   #ifdef B_USE_DATA_QUEUES
   AttemptReceiveQueue(myIncomingQueue); //Read as much as possible from the driver's in-queue.
   nConnectStatus=nNotConnected;
   return WINSOCK_NOERROR;
   #else
   return WINSOCK_NOERROR;
   #endif //B_USE_DATA_QUEUES
}









